from pathlib import Path
from ruamel.yaml import YAML

from freemocap import recordingconfig

class Session: #self like "recording self"
    def __init__(self):
        
        self.numCams = None
        self.numFrames = None
        self.numTrackedPoints = None 
        self.basePath = Path.cwd()
        
        self.sessionID = '' #The sessionID tag will be used to generate files names and whatnot
        self.sessionPath = '' #The folder where the to-be-processed videos live (in a folder called "synced Vids")
        #self.numCams = ''#The number of cameras used in this recording self
        self.openPoseDataPath   = ''#Where the open pose data lives
        self.dlcDataPath    = ''#Where the DLC data lives
        self.openPoseExePath = r'C:\openpose'

        self.cam_inputs = None
        self.parameterDictionary = None
        self.rotationInputs = None

        self.cgroup = None

        self.mean_charuco_fr_mar_dim = None
        self.skel_fr_mar_dim = None

        self.mediaPipe_imgPathList = None
        self.openPose_imgPathList = None

    def start_session(self,paramDict,rotDict):
        #start a session, create all the file paths necessary, and create a session dictionary to save settings
        recordPath = self.basePath/'Data'
        recordPath.mkdir(exist_ok= True)

        self.sessionPath = recordPath/self.sessionID
        self.sessionPath.mkdir(exist_ok=True)

        #create all the session filepaths and settings - create and load them into a dictionary
        self.pathList = self.create_session_paths()
        self.session_settings = self.create_session_dictionary()
        self.create_session_txt(paramDict,rotDict)

        self.save_session()



    def create_session_paths(self):
        #creates Path objects to each folder created in a recording self and adds them all to the pathList variable
        #folders for these paths are created at the start of each stage where they are needed 

        pathList = []

        self.rawVidPath = self.sessionPath/'RawVideos'
        pathList.append(self.rawVidPath)

        self.syncedVidPath = self.sessionPath/'SyncedVideos'
        pathList.append(self.syncedVidPath)

        self.calVidPath = self.sessionPath/'CalVideos'
        pathList.append(self.calVidPath)

        self.dataArrayPath = self.sessionPath/'DataArrays'
        pathList.append(self.dataArrayPath)

        self.dlcDataPath = self.sessionPath/'DLCData'
        pathList.append(self.dlcDataPath)

        self.openPoseDataPath = self.sessionPath/'OpenPoseData'
        pathList.append(self.openPoseDataPath)

        self.mediaPipeDataPath = self.sessionPath/'MediaPipeData'
        pathList.append(self.mediaPipeDataPath)

        self.imOutPath = self.sessionPath/'imOut'
        pathList.append(self.imOutPath)
        
        return pathList

    def create_session_dictionary(self):
        #creates a dictionary of settings that should be saved into a YAML throughout each self

        recording_parameters = {'numCams': self.numCams, 'numFrames': self.numFrames, 'numTrackedPoints': self.numTrackedPoints}

        #create a dictionary with the folder names (taken from the recordingconfig.py file) and the folder paths from pathList
        #NOTE AC- 6/30/21 - as of right now, the order of folder names in the recordingconfig.folder_setup list needs to match the 
        #                the order of folder paths in the pathList
        session_paths = {}

        for foldername, folderpath in zip(recordingconfig.folder_setup,self.pathList):
            session_paths[foldername] = folderpath

        session_settings_dictionary = {'recording_parameters':recording_parameters, 'session_paths':session_paths}
        
        return session_settings_dictionary


    def create_session_txt(self,paramDict,rotDict):
    #create a text file listing recording parameters
        parameter_text = self.sessionPath/'sessionSettings.txt' 
        text = open(parameter_text, 'w')
        text.write("Session ID = %s\n" %(self.sessionID))
        text.write("%s = %s\n" %("Parameters", paramDict))
        text.write("%s = %s\n" %("Rotations", rotDict))
        text.close()

    def save_session(self):
        #save the self settings you want to keep from self to self into a yaml
        #because some items are not yaml-able, a copy of the sessionDictionary is made and adjusted to be fully yaml-able
            #Adjustments Made:
                #pathLib objects are all converted to strings
        session_dictionary_to_save = self.session_settings

        for key,value in session_dictionary_to_save['session_paths'].items():
            session_dictionary_to_save['session_paths'][key] = str(value)

        self.session_yaml_path = self.sessionPath/'{}_config.yaml'.format(self.sessionID)
        
        session_yaml = YAML()
        

        with open(self.session_yaml_path,'w') as outfile:
            session_yaml.dump(session_dictionary_to_save,outfile)

   
    def initialize(self,stage):
        #load all session settings back into the session class for this run-through of the code
        
        recordPath = self.basePath/'Data' #create a Data folder in the filepath if none exists yet
        recordPath.mkdir(exist_ok= True)

        self.sessionPath = recordPath/self.sessionID
        self.sessionPath.mkdir(exist_ok=True)

        self.session_yaml_path = self.sessionPath/'{}_config.yaml'.format(self.sessionID)

        if stage == 3:
            #this is for the case of GoPro recordings/external recordings - if no config file exists, create one
            if self.session_yaml_path.is_file():
                self.session_settings = self.load_session()
            else: 
                self.start_session({},{})
        else:
            self.session_settings = self.load_session()
        


   
    def load_session(self):

        #session_yaml_path = self.sessionPath/'{}_config.yaml'.format(self.sessionID)
        session_yaml = YAML(typ='safe', pure=True)

        with open(self.session_yaml_path,'r') as fp:
            session_dictionary_to_load = session_yaml.load(fp)

        session_settings_dictionary = session_dictionary_to_load #create a copy of the loaded dictionary

        for key,value in session_settings_dictionary['session_paths'].items():
            session_settings_dictionary['session_paths'][key] = Path(value)

        self.load_session_paths(session_settings_dictionary)
        self.numCams = session_settings_dictionary['recording_parameters']['numCams']
        self.numFrames = session_settings_dictionary['recording_parameters']['numFrames']
        self.numTrackedPoints = session_settings_dictionary['recording_parameters']['numTrackedPoints']

        return session_settings_dictionary
        f = 2
    
    def load_session_paths(self, session_settings_dict):

        session_paths = session_settings_dict['session_paths']

        self.rawVidPath = session_paths['RawVideos']
        self.syncedVidPath = session_paths['SyncedVideos']
        self.calVidPath = session_paths['CalVideos']
        self.mediaPipeDataPath = session_paths['MediaPipeData']
        self.openPoseDataPath = session_paths['OpenPoseData']
        self.dlcDataPath = session_paths['DLCData']
        self.imOutPath = session_paths['imOut']
        self.dataArrayPath = session_paths['DataArrays']

        

        f = 2

    
        


        

        
            

    


  





from tqdm import tqdm
import cv2


def TestDevice(source):
    cap = cv2.VideoCapture(source, cv2.CAP_DSHOW)
    # if cap is None or not cap.isOpened():
    # print('Warning: unable to open video source: ', source)

    if cap.isOpened():
        # print('Opened: ',source)
        # print('Exposure: '+ str(cap.get(cv2.CAP_PROP_EXPOSURE)))
        # time.sleep(3)
        cap.release()
        cv2.destroyAllWindows()
        open_cam = source
        return open_cam
    else:
        return None


def CreateAvailableCamList():
    openCamList = []
    for x in tqdm(range(20)):  # range 20 right now to be safe
        openCamera = TestDevice(x)
        if openCamera is not None:
            openCamList.append(openCamera)

    return openCamList



import os
import subprocess
import json

import numpy as np
from rich.progress import track
from ruamel.yaml import YAML
from pathlib import Path

# from numba import jit


def runOpenPose(session, dummyRun=False):

    if not dummyRun:
        os.chdir(session.openPoseExePath)

    openPose_jsonPathList = []  # list to hold the paths to the json files
    openPose_imgPathList = []
    openPose_imgPathList_yaml = []
    openPose_jsonPathList_yaml = []

    for (
        thisVidPath
    ) in (
        session.syncedVidPath.iterdir()
    ):  # Run OpenPose ('Windows Portable Demo') on each video in the raw video folder
        if (
            thisVidPath.suffix == ".mp4"
        ):  # NOTE - build some list of 'synced video names' and check against that
            print("OpenPosing: ", thisVidPath.name)
            vidPath = session.openPoseDataPath / thisVidPath.stem
            jsonPath = vidPath / "json"
            jsonPath.mkdir(
                parents=True, exist_ok=True
            )  # this camera's json files (with keypoints)
            imgPath = vidPath / "img"
            imgPath.mkdir(parents=True, exist_ok=True)
            openPose_jsonPathList.append(jsonPath)
            openPose_imgPathList.append(imgPath)
            openPose_imgPathList_yaml.append(str(imgPath))
            openPose_jsonPathList_yaml.append(str(jsonPath))
            if not dummyRun:
                subprocess.run(
                    [
                        "bin\OpenPoseDemo.exe",
                        "--video",
                        str(thisVidPath),
                        "--write_json",
                        str(jsonPath),
                        "--write_images",
                        str(imgPath),
                        "--write_images_format",
                        "png",
                        "--net_resolution",
                        "-1x320",
                        "--hand",
                        "--face",
                        "--number_people_max",
                        "1",
                    ],
                    shell=True,
                )
        else:
            print("Skipping: ", thisVidPath.name)
    session.openPose_jsonPathList = openPose_jsonPathList
    session.openPose_imgPathList = openPose_imgPathList
    
    session.session_settings['openPose_imgPathList'] = openPose_imgPathList_yaml
    session.session_settings['openPose_jsonPathList'] = openPose_jsonPathList_yaml
    # yaml = YAML()
    # data = yaml.load(session.session_yaml_path)
    # data['openPose_imgPathList'] = openPose_imgPathList_yaml
    # data['openPose_jsonPathList'] = openPose_jsonPathList_yaml
    # yaml.dump(data, session.session_yaml_path)
    
    os.chdir(session.sessionPath)


# @jit(nopython=True)
def parseOpenPose(session):
        thisCamNum = -1

        ## %%
        #build header for dataframe - NOTE - #openpose data comes in a line ordered 'pixel x location (px)', 'pixel y (py)', 'confidence (conf)' for each keypoint  
        
        dataFrameHeader = []
        bodyCols = 75
        handCols = 63 #per hand
        faceCols = 210 #das alotta face!
        headerLength = bodyCols + 2*handCols + faceCols#should be 411 for whatever version of openpose i was using on 11 Jan 2021
        numImgPoints = headerLength

        for bb in range(0,int(bodyCols/3)): #loop through the number of body markers (i.e. #bodyCols/3)
            dataFrameHeader.append('body_' + str(bb).zfill(3) + '_pixx')
            dataFrameHeader.append('body_' + str(bb).zfill(3) + '_pixy')
            dataFrameHeader.append('body_' + str(bb).zfill(3) + '_conf')

        for hr in range(0,int(handCols/3)): #loop through the number of handR markers (i.e. #handCols/3)
            dataFrameHeader.append('handR_' + str(hr).zfill(3) + '_pixx')
            dataFrameHeader.append('handR_' + str(hr).zfill(3) + '_pixy')
            dataFrameHeader.append('handR_' + str(hr).zfill(3) + '_conf')

        for hl in range(0,int(handCols/3)): #loop through the number of handL markers (i.e. #handCols/3)
            dataFrameHeader.append('handL_' + str(hl).zfill(3) + '_pixx')
            dataFrameHeader.append('handL_' + str(hl).zfill(3) + '_pixy')
            dataFrameHeader.append('handL_' + str(hl).zfill(3) + '_conf')

        for ff in range(0,int(faceCols/3)): #loop through the number of Face markers (i.e. #faceCols/3)
            dataFrameHeader.append('face_' + str(ff).zfill(3) + '_pixx')
            dataFrameHeader.append('face_' + str(ff).zfill(3) + '_pixy')
            dataFrameHeader.append('face_' + str(ff).zfill(3) + '_conf')

        assert len(dataFrameHeader) == headerLength, ['Header is the wrong length! Should be ' +  str(headerLength) + ' but it is ' + str(len(dataFrameHeader)) + ' Check version of OpenPose?']

        ## %% 
        ## load in data from json files
        #numFrames = int(len(list(Path(session.openPose_jsonPathList[0]).glob('*')))) #lol
        numFrames = session.numFrames
        numMarkers= int(int(len(dataFrameHeader)/3))
        numCams = int(session.numCams)

        openPoseData_nCams_nFrames_nImgPts_XYC = np.ndarray([numCams,numFrames,numMarkers,3]) #hardcoding for now because I am a bad person

        for thisCams_JsonFolderPath in track(session.openPose_jsonPathList, description='Parsing json\'s into a dataframe (per cam)' ):
            thisCamNum += 1
            # print('Parsing into a dataframe: ', thisCams_JsonFolderPath.name )
            jsonPaths = sorted(Path(thisCams_JsonFolderPath).glob('*.json')) #glob is a "generator(?)" for paths to all the jason for THIS camara            
            
            for thisJsonPath in jsonPaths: #loop throug all the json files and save their 'people' data to a dictionary (which will then be formatted into a pandas dataframe). NOTE - will be empty array if no hoomans visible in frame
                # print('loading: ', thisJsonPath.name)
                frameNum = int(thisJsonPath.stem.split('_')[-2]) #frame number we're on
                thisJsonData = json.loads(thisJsonPath.read_bytes())
                thisJsonData = thisJsonData['people'] # #FEATURE_REQUEST -  at some point, we should check the openpose version (save it with the data somehow, verify everything is the same version, use different markernamlists for different versions, etc)

                if thisJsonData: #if this json has data
                    bodyData  = np.array(thisJsonData[0]['pose_keypoints_2d'])
                    handRData = np.array(thisJsonData[0]['hand_right_keypoints_2d'])
                    handLData = np.array(thisJsonData[0]['hand_left_keypoints_2d'])
                    faceData  = np.array(thisJsonData[0]['face_keypoints_2d'])
                    thisFrameRow = np.hstack((bodyData,handRData, handLData, faceData)) #horizontally concatenate these arrays                
                else: #if this json is empty, just stuff it fulla NaNs
                    thisFrameRow = np.empty([headerLength])
                    thisFrameRow.fill(np.nan)


                assert thisFrameRow.size == headerLength, ['Header is the wrong length! Should be ' +  str(headerLength) + ' but it is ' + str(thisFrameRow.size) + ' Check version of OpenPose?']
                
                if frameNum < openPoseData_nCams_nFrames_nImgPts_XYC.shape[1]: #NOTE: THIS SHOULDN"T BE NECESSARY. CURRENT Vid Trim Methods can (apparently) produce vids with off-by-one numbers of frames
                    openPoseData_nCams_nFrames_nImgPts_XYC[thisCamNum, frameNum, :, :] = np.reshape(thisFrameRow, [137,3]) #hard coding b/c I'm a bad person
                
            
        

        # session.dataFrameHeader = dataFrameHeader
        # session.numImgPoints = numImgPoints
        #session.numFrames = frameNum  #NOTE - Need to find a safer way to get this number

        path_to_openpose_2d = session.dataArrayPath/'openPoseData_2d.npy'
        np.save(path_to_openpose_2d, openPoseData_nCams_nFrames_nImgPts_XYC)

        return openPoseData_nCams_nFrames_nImgPts_XYC



from aniposelib.cameras import CameraGroup
from aniposelib.utils import load_pose2d_fnames
import numpy as np
import cv2

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from freemocap import reconstruct3D, fmc_anipose
import os
from rich.progress import track



def CalibrateCaptureVolume(session,board):
    
    session.calVidPath.mkdir(exist_ok = True)
    session.dataArrayPath.mkdir(exist_ok = True)

    calVideoFrameLength = 60
    createCalibrationVideos(session, calVideoFrameLength)

    vidnames = []
    cam_names = []
    for count, thisVidPath in enumerate(session.calVidPath.glob("*.mp4"), start=1):
        vidnames.append([str(thisVidPath)])
        cam_names.append(str(count))
        session.numCams = count

    cgroup = fmc_anipose.CameraGroup.from_names(
        cam_names, fisheye=True
    )  # Looking through their code... it looks lke the 'fisheye=True' doesn't do much (see 2020-03-29 obsidian note)

    calibrationFile = "{}_calibration.yaml".format(session.sessionID)

    session.cameraCalFilePath = session.sessionPath / calibrationFile

    forceRunCameraExtrinsicsCalibration = True  # set this to True to re-run the camera calibration and saveover the existing config file

    # if not session.cameraCalFilePath.exists() or forceRunCameraExtrinsicsCalibration: #if that config file doesn't exist, run the camera calibration whosists. If it does, then just load the toml
    ## this will take a few minutes
    ## it will detect the charuco board in the videos,
    ## then calibrate the cameras based on the detections, using iterative bundle adjustment
    # error,all_rows,merged = cgroup.calibrate_videos(vidnames, board) #JSM NOTE - in its original form, this method doesn't throw an error when a video fails to load. This is an easy fix!

    #%% This next bit is a copy-paste of the inner bits of `cgroup.calibrate_videos` - Part 1 of 2
    """Takes as input a list of list of video filenames, one list of each camera.
    Also takes a board which specifies what should be detected in the videos"""

    all_rows = cgroup.get_rows_videos(vidnames, board, verbose=True)

    cgroup.set_camera_sizes_videos(vidnames)

    if session.debug:
        fig = plt.figure(47290)
        for camNum in range(len(all_rows)):
            ax = fig.add_subplot(2, 2, camNum + 1)
            # cap = cv2.VideoCapture(vidnames[camNum][0]) #this is crazy inefficient - we should save out the first image somewhere and save it in an accessible location
            # ret, frame = cap.read()
            # plt.imshow(frame)
            # cap.release()
            for frNum in range(len(all_rows[camNum])):
                corners = np.squeeze(all_rows[camNum][frNum]["filled"])
                plt.plot(corners[:, 0], corners[:, 1], ".-")
            # xmin = np.nanmin(corners[:,0])*.9
            # xmax = np.nanmax(corners[:,0])*1.1
            # ymin = np.nanmin(corners[:,1])*.9
            # ymax = np.nanmax(corners[:,1])*1.1
            # ax.set_xlim(xmin, xmax)
            # ax.set_xlim(ymin, ymax)
        plt.show()

    #%% This next bit is a copy-paste of the inner bits of `cgroup.calibrate_videos` - Part 2 of 2
    # error, merged = cgroup.calibrate_rows(all_rows, board,
    #                            init_intrinsics=True,
    #                            init_extrinsics=True)
    #%% End of the copy-paste of the inner bits of `cgroup.calibrate_videos` - END

    # JSM NOTE - need add method to extract Charuco board points from the calibrate_videos pipeline

    ## if you need to save and load
    ## example saving and loading for later
    if not session.cameraCalFilePath.is_file():
        error, merged = cgroup.calibrate_videos(vidnames, board)
        cgroup.dump(
            session.cameraCalFilePath
        )  # JSM NOTE  - let's just use .yaml's unless there is some reason to use .toml
        mergename = session.sessionPath / "merged.npy"
        np.save(mergename, merged)
    else:
        cgroup = CameraGroup.load(
            session.cameraCalFilePath
        )  # load previous calibration config

    #JSM NOTE - need add method to extract Charuco board points from the calibrate_videos pipeline

        ## if you need to save and load
        ## example saving and loading for later
    if not session.cameraCalFilePath.is_file(): 
        error,merged = cgroup.calibrate_videos(vidnames, board)
        cgroup.dump(session.cameraCalFilePath) #JSM NOTE  - let's just use .yaml's unless there is some reason to use .toml
        mergename = session.dataArrayPath/'charuco_2d_points.npy'
        np.save(mergename,merged)
    else: 
        cgroup = CameraGroup.load(session.cameraCalFilePath) #load previous calibration config


    session.cgroup = cgroup
    n_frames = 40
    startframe = 0
    n_trackedPoints = 24
    framelist = range(startframe, startframe + n_frames)

    charucoarray = np.empty([session.numCams, n_frames, n_trackedPoints, 1, 2])
    charucoarray[:] = np.nan

    data = np.load(session.dataArrayPath/'charuco_2d_points.npy',allow_pickle = True)

    for cam in range(session.numCams):
        for count, frame in enumerate(framelist):
            try:
                charucoarray[cam][count] = data[frame][cam]["filled"]
            except:
                print("failed frame:", frame)
                continue
    charuco_nCams_nFrames_nImgPts_XY = np.squeeze(np.array(charucoarray))

    session.charuco_nCams_nFrames_nImgPts_XY = charuco_nCams_nFrames_nImgPts_XY

    charuco_fr_mar_dim = reconstruct3D.reconstruct3D(
        session, charuco_nCams_nFrames_nImgPts_XY
    )

    mean_charuco_fr_mar_dim = np.nanmean(charuco_fr_mar_dim, axis=0)

    # charry_reshaped = charucoarray.reshape(session.numCams, -1, 2)

    # char_flat = cgroup.triangulate(charry_reshaped, progress=True)
    # charReprojerr_flat = cgroup.reprojection_error(char_flat, charry_reshaped, mean=True)

    # char_fr_mar_dim = char_flat.reshape(n_frames, n_trackedPoints, 3)
    # charReprojErr_fr_mar_err = charReprojerr_flat.reshape(n_frames, n_trackedPoints)

    if session.debug:
        fig = plt.figure()
        # mean charuco position
        ax1 = fig.add_subplot(111, projection="3d")
        ax1.cla()
        x = mean_charuco_fr_mar_dim[:][:, 0]
        y = mean_charuco_fr_mar_dim[:][:, 1]
        z = mean_charuco_fr_mar_dim[:][:, 2]
        mx = np.nanmean(x)
        my = np.nanmean(y)
        mz = np.nanmean(z)

        ax1.scatter(x, y, z, marker="o")
        ax1.set_title("Mean Charuco Point Positions")

        ax1.set_xlabel("x")
        ax1.set_ylabel("y")
        ax1.set_zlabel("z")

        for camNum in range(len(cgroup.cameras)):
            np.append(mx, cgroup.cameras[camNum].tvec[0])
            np.append(my, cgroup.cameras[camNum].tvec[1])
            np.append(mz, cgroup.cameras[camNum].tvec[2])

            ax1.scatter(
                cgroup.cameras[camNum].tvec[0],
                cgroup.cameras[camNum].tvec[1],
                cgroup.cameras[camNum].tvec[2],
                marker="p",
            )

        axRange = board.square_length * 5

        ax1.set_xlim(mx - axRange, mx + axRange)
        ax1.set_ylim(my - axRange, my + axRange)
        ax1.set_zlim(mz - axRange, mz + axRange)

        # #charuco points over time
        # ax2 = fig.add_subplot(122)
        # ax2.cla()
        # numPts = charuco_fr_mar_dim.shape[1]
        # for pp in range(numPts):
        #     ax2.plot(charuco_fr_mar_dim[:,pp,:])
        # ax2.set_title('Charuco Point positions over time')
        plt.show()

    path_to_charuco_array = session.dataArrayPath/'charuco_3d_points.npy'
    np.save(path_to_charuco_array, mean_charuco_fr_mar_dim)
    return cgroup, mean_charuco_fr_mar_dim


def createCalibrationVideos(session, calVideoFrameLength):
    vidList = os.listdir(session.syncedVidPath)
    framelist = list(range(calVideoFrameLength))
    codec = "DIVX"
    for count, vid in enumerate(vidList, start=1):
        cam_name = "Cam{}".format(count)
        cap = cv2.VideoCapture(str(session.syncedVidPath / vid))
        fourcc = cv2.VideoWriter_fourcc(*codec)

        # grab resolution parameters from the videos
        resWidth = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
        resHeight = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
        framerate = int(cap.get(cv2.CAP_PROP_FPS))

        saveName = (
            session.sessionID + "_trimmed_" + cam_name + ".mp4"
        )  # create a name for the trimmed video
        saveCalVidPath = str(
            session.calVidPath / saveName
        )  # create an output path for the function

        success, image = cap.read()  # start reading frames

        out = cv2.VideoWriter(saveCalVidPath, fourcc, framerate, (resWidth, resHeight))
        print("Trimming " + cam_name)
        for frame in track(framelist):
            cap.set(
                cv2.CAP_PROP_POS_FRAMES, frame
            )  # set the video to the frame that we need
            success, image = cap.read()
            out.write(image)
        cap.release()
        out.release()
